# type: ignore
import json
import os
import subprocess
import uuid
import zipfile
from string import Template

import click

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from ..telemetry import track
from ._utils._console import ConsoleLogger

console = ConsoleLogger()

schema = "https://cloud.uipath.com/draft/2024-12/entry-point"


def validate_config_structure(config_data):
    required_fields = ["entryPoints"]
    for field in required_fields:
        if field not in config_data:
            console.error(f"uipath.json is missing the required field: {field}.")


def check_config(directory):
    config_path = os.path.join(directory, "uipath.json")
    toml_path = os.path.join(directory, "pyproject.toml")

    if not os.path.isfile(config_path):
        console.error("uipath.json not found, please run `uipath init`.")
    if not os.path.isfile(toml_path):
        console.error("pyproject.toml not found.")

    with open(config_path, "r") as config_file:
        config_data = json.load(config_file)

    validate_config_structure(config_data)

    toml_data = read_toml_project(toml_path)

    return {
        "project_name": toml_data["name"],
        "description": toml_data["description"],
        "entryPoints": config_data["entryPoints"],
        "version": toml_data["version"],
        "authors": toml_data["authors"],
    }


def is_uv_available():
    """Check if uv command is available in the system."""
    try:
        subprocess.run(["uv", "--version"], capture_output=True, check=True, timeout=20)
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False
    except Exception as e:
        console.warning(
            f"An unexpected error occurred while checking uv availability: {str(e)}"
        )
        return False


def is_uv_project(directory):
    """Check if this is a uv project by looking for the uv.lock file."""
    uv_lock_path = os.path.join(directory, "uv.lock")

    # If uv.lock exists, it's definitely a uv project
    if os.path.exists(uv_lock_path):
        return True

    return False


def run_uv_lock(directory):
    """Run uv lock to update the lock file."""
    try:
        subprocess.run(
            ["uv", "lock"],
            cwd=directory,
            capture_output=True,
            text=True,
            check=True,
            timeout=60,
        )
        return True
    except subprocess.CalledProcessError as e:
        console.warning(f"uv lock failed: {e.stderr}")
        return False
    except FileNotFoundError:
        console.warning("uv command not found. Skipping lock file update.")
        return False
    except Exception as e:
        console.warning(f"An unexpected error occurred while running uv lock: {str(e)}")
        return False


def handle_uv_operations(directory):
    """Handle uv operations if uv is detected and available."""
    if not is_uv_available():
        return

    if not is_uv_project(directory):
        return

    # Always run uv lock to ensure lock file is up to date
    run_uv_lock(directory)


def generate_operate_file(entryPoints):
    project_id = str(uuid.uuid4())

    first_entry = entryPoints[0]
    file_path = first_entry["filePath"]
    type = first_entry["type"]

    operate_json_data = {
        "$schema": schema,
        "projectId": project_id,
        "main": file_path,
        "contentType": type,
        "targetFramework": "Portable",
        "targetRuntime": "python",
        "runtimeOptions": {"requiresUserInteraction": False, "isAttended": False},
    }

    return operate_json_data


def generate_entrypoints_file(entryPoints):
    entrypoint_json_data = {
        "$schema": schema,
        "$id": "entry-points.json",
        "entryPoints": entryPoints,
    }

    return entrypoint_json_data


def generate_bindings_content():
    bindings_content = {"version": "2.0", "resources": []}

    return bindings_content


def get_proposed_version(directory):
    output_dir = os.path.join(directory, ".uipath")
    if not os.path.exists(output_dir):
        return None

    # Get all .nupkg files
    nupkg_files = [f for f in os.listdir(output_dir) if f.endswith(".nupkg")]
    if not nupkg_files:
        return None

    # Sort by modification time to get most recent
    latest_file = max(
        nupkg_files, key=lambda f: os.path.getmtime(os.path.join(output_dir, f))
    )

    # Extract version from filename
    # Remove .nupkg extension first
    name_version = latest_file[:-6]
    # Find 3rd last occurrence of . by splitting and joining parts
    parts = name_version.split(".")
    if len(parts) >= 3:
        version = ".".join(parts[-3:])
    else:
        version = name_version

    # Increment patch version by 1
    try:
        major, minor, patch = version.split(".")
        new_version = f"{major}.{minor}.{int(patch) + 1}"
        return new_version
    except Exception:
        return "0.0.1"


def generate_content_types_content():
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", "[Content_Types].xml.template"
    )
    with open(templates_path, "r") as file:
        content_types_content = file.read()
    return content_types_content


def generate_nuspec_content(projectName, packageVersion, description, authors):
    variables = {
        "packageName": projectName,
        "packageVersion": packageVersion,
        "description": description,
        "authors": authors,
    }
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", "package.nuspec.template"
    )
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()
    return Template(content).substitute(variables)


def generate_rels_content(nuspecPath, psmdcpPath):
    # /package/services/metadata/core-properties/254324ccede240e093a925f0231429a0.psmdcp
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", ".rels.template"
    )
    nuspecId = "R" + str(uuid.uuid4()).replace("-", "")[:16]
    psmdcpId = "R" + str(uuid.uuid4()).replace("-", "")[:16]
    variables = {
        "nuspecPath": nuspecPath,
        "nuspecId": nuspecId,
        "psmdcpPath": psmdcpPath,
        "psmdcpId": psmdcpId,
    }
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()
    return Template(content).substitute(variables)


def generate_psmdcp_content(projectName, version, description, authors):
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", ".psmdcp.template"
    )

    token = str(uuid.uuid4()).replace("-", "")[:32]
    random_file_name = f"{uuid.uuid4().hex[:16]}.psmdcp"
    variables = {
        "creator": authors,
        "description": description,
        "packageVersion": version,
        "projectName": projectName,
        "publicKeyToken": token,
    }
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()

    return [random_file_name, Template(content).substitute(variables)]


def generate_package_descriptor_content(entryPoints):
    files = {
        "operate.json": "content/operate.json",
        "entry-points.json": "content/entry-points.json",
        "bindings.json": "content/bindings_v2.json",
    }

    for entry in entryPoints:
        files[entry["filePath"]] = entry["filePath"]

    package_descriptor_content = {
        "$schema": "https://cloud.uipath.com/draft/2024-12/package-descriptor",
        "files": files,
    }

    return package_descriptor_content


def is_venv_dir(d):
    return (
        os.path.exists(os.path.join(d, "Scripts", "activate"))
        if os.name == "nt"
        else os.path.exists(os.path.join(d, "bin", "activate"))
    )


def pack_fn(
    projectName,
    description,
    entryPoints,
    version,
    authors,
    directory,
    include_uv_lock=True,
):
    operate_file = generate_operate_file(entryPoints)
    entrypoints_file = generate_entrypoints_file(entryPoints)

    # Get bindings from uipath.json if available
    config_path = os.path.join(directory, "uipath.json")
    if not os.path.exists(config_path):
        console.error("uipath.json not found, please run `uipath init`.")

    # Define the allowlist of file extensions to include
    file_extensions_included = [".py", ".mermaid", ".json", ".yaml", ".yml"]
    files_included = []
    # Binary files that should be read in binary mode
    binary_extensions = [".exe", "", ".xlsx", ".xls"]

    with open(config_path, "r") as f:
        config_data = json.load(f)
        if "bindings" in config_data:
            bindings_content = config_data["bindings"]
        else:
            bindings_content = generate_bindings_content()
        if "settings" in config_data:
            settings = config_data["settings"]
            if "fileExtensionsIncluded" in settings:
                file_extensions_included.extend(settings["fileExtensionsIncluded"])
            if "filesIncluded" in settings:
                files_included = settings["filesIncluded"]

    content_types_content = generate_content_types_content()
    [psmdcp_file_name, psmdcp_content] = generate_psmdcp_content(
        projectName, version, description, authors
    )
    nuspec_content = generate_nuspec_content(projectName, version, description, authors)
    rels_content = generate_rels_content(
        f"/{projectName}.nuspec",
        f"/package/services/metadata/core-properties/{psmdcp_file_name}",
    )
    package_descriptor_content = generate_package_descriptor_content(entryPoints)

    # Create .uipath directory if it doesn't exist
    os.makedirs(".uipath", exist_ok=True)

    with zipfile.ZipFile(
        f".uipath/{projectName}.{version}.nupkg", "w", zipfile.ZIP_DEFLATED
    ) as z:
        # Add metadata files
        z.writestr(
            f"./package/services/metadata/core-properties/{psmdcp_file_name}",
            psmdcp_content,
        )
        z.writestr("[Content_Types].xml", content_types_content)
        z.writestr(
            "content/package-descriptor.json",
            json.dumps(package_descriptor_content, indent=4),
        )
        z.writestr("content/operate.json", json.dumps(operate_file, indent=4))
        z.writestr("content/entry-points.json", json.dumps(entrypoints_file, indent=4))
        z.writestr("content/bindings_v2.json", json.dumps(bindings_content, indent=4))
        z.writestr(f"{projectName}.nuspec", nuspec_content)
        z.writestr("_rels/.rels", rels_content)

        # Walk through directory and add all files with extensions in the allowlist
        for root, dirs, files in os.walk(directory):
            # Skip all directories that start with . or are a venv
            dirs[:] = [
                d
                for d in dirs
                if not d.startswith(".") and not is_venv_dir(os.path.join(root, d))
            ]

            for file in files:
                file_extension = os.path.splitext(file)[1].lower()
                if file_extension in file_extensions_included or file in files_included:
                    file_path = os.path.join(root, file)
                    rel_path = os.path.relpath(file_path, directory)
                    if file_extension in binary_extensions:
                        # Read binary files in binary mode
                        with open(file_path, "rb") as f:
                            z.writestr(f"content/{rel_path}", f.read())
                    else:
                        try:
                            # Try UTF-8 first
                            with open(file_path, "r", encoding="utf-8") as f:
                                z.writestr(f"content/{rel_path}", f.read())
                        except UnicodeDecodeError:
                            # If UTF-8 fails, try with utf-8-sig (for files with BOM)
                            try:
                                with open(file_path, "r", encoding="utf-8-sig") as f:
                                    z.writestr(f"content/{rel_path}", f.read())
                            except UnicodeDecodeError:
                                # If that also fails, try with latin-1 as a fallback
                                with open(file_path, "r", encoding="latin-1") as f:
                                    z.writestr(f"content/{rel_path}", f.read())

        # Handle optional files, conditionally including uv.lock
        optional_files = ["pyproject.toml", "README.md"]
        if include_uv_lock:
            optional_files.append("uv.lock")

        for file in optional_files:
            file_path = os.path.join(directory, file)
            if os.path.exists(file_path):
                try:
                    with open(file_path, "r", encoding="utf-8") as f:
                        z.writestr(f"content/{file}", f.read())
                except UnicodeDecodeError:
                    with open(file_path, "r", encoding="latin-1") as f:
                        z.writestr(f"content/{file}", f.read())


def read_toml_project(file_path: str) -> dict[str, any]:
    with open(file_path, "rb") as f:
        content = tomllib.load(f)
        if "project" not in content:
            console.error("pyproject.toml is missing the required field: project.")
        if "name" not in content["project"]:
            console.error("pyproject.toml is missing the required field: project.name.")
        if "description" not in content["project"]:
            console.error(
                "pyproject.toml is missing the required field: project.description."
            )
        if "version" not in content["project"]:
            console.error(
                "pyproject.toml is missing the required field: project.version."
            )

        return {
            "name": content["project"]["name"],
            "description": content["project"]["description"],
            "version": content["project"]["version"],
            "authors": content["project"].get("authors", [{"name": ""}])[0]["name"],
        }


def get_project_version(directory):
    toml_path = os.path.join(directory, "pyproject.toml")
    if not os.path.exists(toml_path):
        console.warning("pyproject.toml not found. Using default version 0.0.1")
        return "0.0.1"
    toml_data = read_toml_project(toml_path)
    return toml_data["version"]


def display_project_info(config):
    max_label_length = max(
        len(label) for label in ["Name", "Version", "Description", "Authors"]
    )

    max_length = 100
    description = config["description"]
    if len(description) >= max_length:
        description = description[: max_length - 3] + " ..."

    console.log(f"{'Name'.ljust(max_label_length)}: {config['project_name']}")
    console.log(f"{'Version'.ljust(max_label_length)}: {config['version']}")
    console.log(f"{'Description'.ljust(max_label_length)}: {description}")
    console.log(f"{'Authors'.ljust(max_label_length)}: {config['authors']}")


@click.command()
@click.argument("root", type=str, default="./")
@click.option(
    "--nolock",
    is_flag=True,
    help="Skip running uv lock and exclude uv.lock from the package",
)
@track
def pack(root, nolock):
    """Pack the project."""
    version = get_project_version(root)

    while not os.path.isfile(os.path.join(root, "uipath.json")):
        console.error(
            "uipath.json not found. Please run `uipath init` in the project directory."
        )
    config = check_config(root)
    if not config["project_name"] or config["project_name"].strip() == "":
        console.error(
            "Project name cannot be empty. Please specify a name in pyproject.toml."
        )

    if not config["description"] or config["description"].strip() == "":
        console.error(
            "Project description cannot be empty. Please specify a description in pyproject.toml."
        )

    if not config["authors"] or config["authors"].strip() == "":
        console.error(
            'Project authors cannot be empty. Please specify authors in pyproject.toml:\n    authors = [{ name = "John Doe" }]'
        )

    invalid_chars = ["&", "<", ">", '"', "'", ";"]
    for char in invalid_chars:
        if char in config["project_name"]:
            console.error(f"Project name contains invalid character: '{char}'")

    for char in invalid_chars:
        if char in config["description"]:
            console.error(f"Project description contains invalid character: '{char}'")

    with console.spinner("Packaging project ..."):
        try:
            # Handle uv operations before packaging, unless nolock is specified
            if not nolock:
                handle_uv_operations(root)

            pack_fn(
                config["project_name"],
                config["description"],
                config["entryPoints"],
                version or config["version"],
                config["authors"],
                root,
                include_uv_lock=not nolock,
            )
            display_project_info(config)
            console.success("Project successfully packaged.")

        except Exception as e:
            console.error(
                f"Failed to create package {config['project_name']}.{version or config['version']}: {str(e)}"
            )


if __name__ == "__main__":
    pack()



import json
from unittest.mock import MagicMock, patch

import pytest

from uipath._config import Config
from uipath._execution_context import ExecutionContext
from uipath._services.llm_gateway_service import (
    EmbeddingModels,
    UiPathOpenAIService,
)
from uipath.models.llm_gateway import (
    TextEmbedding,
    UsageInfo,
)


class TestOpenAIService:
    @pytest.fixture
    def config(self):
        return Config(base_url="https://example.com", secret="test_secret")

    @pytest.fixture
    def execution_context(self):
        return ExecutionContext()

    @pytest.fixture
    def openai_service(self, config, execution_context):
        return UiPathOpenAIService(config=config, execution_context=execution_context)

    def test_init(self, config, execution_context):
        service = UiPathOpenAIService(
            config=config, execution_context=execution_context
        )
        assert service._config == config
        assert service._execution_context == execution_context

    @patch.object(UiPathOpenAIService, "request_async")
    @pytest.mark.asyncio
    async def test_embeddings_usage(self, mock_request, openai_service):
        # Mock response
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "encoding": "cl100k_base",
            "prompt_tokens": 4,
        }
        mock_request.return_value = mock_response

        # Call the method
        result = await openai_service.embeddings_usage(input="Test input")

        # Assertions
        mock_request.assert_called_once()
        assert isinstance(result, UsageInfo)
        assert result.encoding == "cl100k_base"
        assert result.prompt_tokens == 4

    @patch.object(UiPathOpenAIService, "request_async")
    @pytest.mark.asyncio
    async def test_embeddings(self, mock_request, openai_service):
        # Mock response
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [{"embedding": [0.1, 0.2, 0.3], "index": 0, "object": "embedding"}],
            "model": "text-embedding-ada-002",
            "object": "list",
            "usage": {"prompt_tokens": 4, "total_tokens": 4},
        }
        mock_request.return_value = mock_response

        # Call the method
        result = await openai_service.embeddings(input="Test input")

        # Assertions
        mock_request.assert_called_once()
        assert isinstance(result, TextEmbedding)
        assert result.data[0].embedding == [0.1, 0.2, 0.3]
        assert result.model == "text-embedding-ada-002"
        assert result.usage.prompt_tokens == 4

    @patch.object(UiPathOpenAIService, "request_async")
    @pytest.mark.asyncio
    async def test_chat_completions_usage(self, mock_request, openai_service):
        # Mock response
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "encoding": "cl100k_base",
            "prompt_tokens": 10,
        }
        mock_request.return_value = mock_response

        # Test messages
        messages = [
            {"role": "system", "content": "You are a helpful assistant"},
            {"role": "user", "content": "Hello"},
        ]

        # Call the method
        result = await openai_service.chat_completions_usage(
            messages=messages, max_tokens=50, temperature=0.5
        )

        # Assertions
        mock_request.assert_called_once()
        assert isinstance(result, UsageInfo)
        assert result.encoding == "cl100k_base"
        assert result.prompt_tokens == 10

        # Verify the correct endpoint was called
        args, kwargs = mock_request.call_args
        assert json.loads(kwargs["content"])["messages"] == messages
        assert json.loads(kwargs["content"])["max_tokens"] == 50
        assert json.loads(kwargs["content"])["temperature"] == 0.5

    @patch.object(UiPathOpenAIService, "request_async")
    @pytest.mark.asyncio
    async def test_embeddings_with_custom_model(self, mock_request, openai_service):
        # Mock response
        mock_response = MagicMock()
        mock_response.json.return_value = {
            "data": [{"embedding": [0.1, 0.2, 0.3], "index": 0, "object": "embedding"}],
            "model": "text-embedding-3-large",
            "object": "list",
            "usage": {"prompt_tokens": 4, "total_tokens": 4},
        }
        mock_request.return_value = mock_response

        # Call the method with custom model
        result = await openai_service.embeddings(
            input="Test input", embedding_model=EmbeddingModels.text_embedding_3_large
        )

        # Assertions for the result
        mock_request.assert_called_once()
        assert result.model == "text-embedding-3-large"
        assert len(result.data) == 1
        assert result.data[0].embedding == [0.1, 0.2, 0.3]
        assert result.data[0].index == 0
        assert result.object == "list"
        assert result.usage.prompt_tokens == 4
        assert result.usage.total_tokens == 4



import functools
import inspect
from typing import Any, Callable, TypeVar

from ._read_overwrites import read_resource_overwrites

T = TypeVar("T")


def infer_bindings(
    resource_type: str, name: str = "name", folder_path: str = "folder_path"
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # convert both args and kwargs to single dict
            sig = inspect.signature(func)
            bound = sig.bind_partial(*args, **kwargs)
            bound.apply_defaults()
            all_args = dict(bound.arguments)

            if name in all_args or folder_path in all_args:
                with read_resource_overwrites(
                    resource_type,
                    all_args.get(name),  # type: ignore
                    all_args.get(folder_path, None),
                ) as (name_overwrite_or_default, folder_path_overwrite_or_default):
                    all_args[name] = name_overwrite_or_default
                    all_args[folder_path] = folder_path_overwrite_or_default

            return func(**all_args)

        wrapper._should_infer_bindings = True  # type: ignore
        wrapper._infer_bindings_mappings = {"name": name, "folder_path": folder_path}  # type: ignore
        return wrapper

    return decorator


def get_inferred_bindings_names(cls: T):
    inferred_bindings = {}
    for name, method in inspect.getmembers(cls, inspect.isfunction):
        if hasattr(method, "_should_infer_bindings") and method._should_infer_bindings:
            inferred_bindings[name] = method._infer_bindings_mappings  # type: ignore

    return inferred_bindings



import os
from unittest.mock import MagicMock, patch

import pytest
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExportResult

from uipath.tracing._otel_exporters import LlmOpsHttpExporter


@pytest.fixture
def mock_env_vars():
    """Fixture to set and clean up environment variables for testing."""
    original_values = {}

    # Save original values
    for var in ["UIPATH_URL", "UIPATH_ACCESS_TOKEN"]:
        original_values[var] = os.environ.get(var)

    # Set test values
    os.environ["UIPATH_URL"] = "https://test.uipath.com/org/tenant/"
    os.environ["UIPATH_ACCESS_TOKEN"] = "test-token"

    yield

    # Restore original values
    for var, value in original_values.items():
        if value is None:
            if var in os.environ:
                del os.environ[var]
        else:
            os.environ[var] = value


@pytest.fixture
def mock_span():
    """Create a mock ReadableSpan for testing."""
    span = MagicMock(spec=ReadableSpan)
    return span


@pytest.fixture
def exporter(mock_env_vars):
    """Create an exporter instance for testing."""
    with patch("uipath.tracing._otel_exporters.Client"):
        exporter = LlmOpsHttpExporter()
        # Mock _build_url to include query parameters as in the actual implementation
        exporter._build_url = MagicMock(  # type: ignore
            return_value="https://test.uipath.com/org/tenant/llmopstenant_/api/Traces/spans?traceId=test-trace-id&source=Robots"
        )
        yield exporter


def test_init_with_env_vars(mock_env_vars):
    """Test initialization with environment variables."""
    with patch("uipath.tracing._otel_exporters.Client"):
        exporter = LlmOpsHttpExporter()

        assert exporter.base_url == "https://test.uipath.com/org/tenant"
        assert exporter.auth_token == "test-token"
        assert exporter.headers == {
            "Content-Type": "application/json",
            "Authorization": "Bearer test-token",
        }


def test_init_with_default_url():
    """Test initialization with default URL when environment variable is not set."""
    with (
        patch("uipath.tracing._otel_exporters.Client"),
        patch.dict(os.environ, {"UIPATH_ACCESS_TOKEN": "test-token"}, clear=True),
    ):
        exporter = LlmOpsHttpExporter()

        assert exporter.base_url == "https://cloud.uipath.com/dummyOrg/dummyTennant"
        assert exporter.auth_token == "test-token"


def test_export_success(exporter, mock_span):
    """Test successful export of spans."""
    mock_uipath_span = MagicMock()
    mock_uipath_span.to_dict.return_value = {"span": "data", "TraceId": "test-trace-id"}

    with patch(
        "uipath.tracing._otel_exporters._SpanUtils.otel_span_to_uipath_span",
        return_value=mock_uipath_span,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 200
        exporter.http_client.post.return_value = mock_response

        result = exporter.export([mock_span])

        assert result == SpanExportResult.SUCCESS
        exporter._build_url.assert_called_once_with(
            [{"span": "data", "TraceId": "test-trace-id"}]
        )
        exporter.http_client.post.assert_called_once_with(
            "https://test.uipath.com/org/tenant/llmopstenant_/api/Traces/spans?traceId=test-trace-id&source=Robots",
            json=[{"span": "data", "TraceId": "test-trace-id"}],
        )


def test_export_failure(exporter, mock_span):
    """Test export failure with multiple retries."""
    mock_uipath_span = MagicMock()
    mock_uipath_span.to_dict.return_value = {"span": "data"}

    with patch(
        "uipath.tracing._otel_exporters._SpanUtils.otel_span_to_uipath_span",
        return_value=mock_uipath_span,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.text = "Internal Server Error"
        exporter.http_client.post.return_value = mock_response

        with patch("uipath.tracing._otel_exporters.time.sleep") as mock_sleep:
            result = exporter.export([mock_span])

        assert result == SpanExportResult.FAILURE
        assert exporter.http_client.post.call_count == 4  # Default max_retries is 3
        assert (
            mock_sleep.call_count == 3
        )  # Should sleep between retries (except after the last one)


def test_export_exception(exporter, mock_span):
    """Test export with exceptions during HTTP request."""
    mock_uipath_span = MagicMock()
    mock_uipath_span.to_dict.return_value = {"span": "data"}

    with patch(
        "uipath.tracing._otel_exporters._SpanUtils.otel_span_to_uipath_span",
        return_value=mock_uipath_span,
    ):
        exporter.http_client.post.side_effect = Exception("Connection error")

        with patch("uipath.tracing._otel_exporters.time.sleep"):
            result = exporter.export([mock_span])

        assert result == SpanExportResult.FAILURE
        assert exporter.http_client.post.call_count == 4  # Default max_retries is 3


def test_force_flush(exporter):
    """Test force_flush returns True."""
    assert exporter.force_flush() is True


def test_get_base_url():
    """Test _get_base_url method with different environment configurations."""
    # Test with environment variable set
    with patch.dict(
        os.environ, {"UIPATH_URL": "https://custom.uipath.com/org/tenant/"}, clear=True
    ):
        with patch("uipath.tracing._otel_exporters.Client"):
            exporter = LlmOpsHttpExporter()
            assert exporter.base_url == "https://custom.uipath.com/org/tenant"

    # Test with environment variable set but with no trailing slash
    with patch.dict(
        os.environ, {"UIPATH_URL": "https://custom.uipath.com/org/tenant"}, clear=True
    ):
        with patch("uipath.tracing._otel_exporters.Client"):
            exporter = LlmOpsHttpExporter()
            assert exporter.base_url == "https://custom.uipath.com/org/tenant"

    # Test with no environment variable
    with patch.dict(os.environ, {}, clear=True):
        with patch("uipath.tracing._otel_exporters.Client"):
            exporter = LlmOpsHttpExporter()
            assert exporter.base_url == "https://cloud.uipath.com/dummyOrg/dummyTennant"


def test_send_with_retries_success():
    """Test _send_with_retries method with successful response."""
    with patch("uipath.tracing._otel_exporters.Client"):
        exporter = LlmOpsHttpExporter()

        mock_response = MagicMock()
        mock_response.status_code = 200
        exporter.http_client.post.return_value = mock_response  # type: ignore

        result = exporter._send_with_retries("http://example.com", [{"span": "data"}])

        assert result == SpanExportResult.SUCCESS
        exporter.http_client.post.assert_called_once_with(  # type: ignore
            "http://example.com", json=[{"span": "data"}]
        )



from typing import Optional


class IngestionInProgressException(Exception):
    """An exception that is triggered when a search is attempted on an index that is currently undergoing ingestion."""

    def __init__(self, index_name: Optional[str], search_operation: bool = True):
        index_name = index_name or "Unknown index name"
        if search_operation:
            self.message = f"index '{index_name}' cannot be searched during ingestion"
        else:
            self.message = f"index '{index_name}' is currently queued for ingestion"
        super().__init__(self.message)



import base64
import hashlib
import json
import os
from urllib.parse import urlencode

from ._models import AuthConfig


def generate_code_verifier_and_challenge():
    """Generate PKCE code verifier and challenge."""
    code_verifier = base64.urlsafe_b64encode(os.urandom(32)).decode("utf-8").rstrip("=")

    code_challenge_bytes = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = (
        base64.urlsafe_b64encode(code_challenge_bytes).decode("utf-8").rstrip("=")
    )

    return code_verifier, code_challenge


def get_state_param() -> str:
    return base64.urlsafe_b64encode(os.urandom(32)).decode("utf-8").rstrip("=")


def get_auth_config() -> AuthConfig:
    auth_config = {}
    with open(os.path.join(os.path.dirname(__file__), "auth_config.json"), "r") as f:
        auth_config = json.load(f)

    port = auth_config.get("port", 8104)

    redirect_uri = auth_config["redirect_uri"].replace("__PY_REPLACE_PORT__", str(port))

    return AuthConfig(
        client_id=auth_config["client_id"],
        redirect_uri=redirect_uri,
        scope=auth_config["scope"],
        port=port,
    )


def get_auth_url(domain: str) -> tuple[str, str, str]:
    """Get the authorization URL for OAuth2 PKCE flow.

    Args:
        domain (str): The UiPath domain to authenticate against (e.g. 'alpha', 'cloud')

    Returns:
        tuple[str, str]: A tuple containing:
            - The authorization URL with query parameters
            - The code verifier for PKCE flow
    """
    code_verifier, code_challenge = generate_code_verifier_and_challenge()
    auth_config = get_auth_config()
    state = get_state_param()
    query_params = {
        "client_id": auth_config["client_id"],
        "redirect_uri": auth_config["redirect_uri"],
        "response_type": "code",
        "scope": auth_config["scope"],
        "state": state,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
    }

    query_string = urlencode(query_params)
    url = f"https://{domain}.uipath.com/identity_/connect/authorize?{query_string}"
    return url, code_verifier, state



import os
import uuid
from json import dumps
from typing import Any, Dict, List, Optional, Tuple

from .._config import Config
from .._execution_context import ExecutionContext
from .._folder_context import FolderContext
from .._utils import Endpoint, RequestSpec
from .._utils.constants import (
    ENV_TENANT_ID,
    HEADER_FOLDER_KEY,
    HEADER_FOLDER_PATH,
    HEADER_TENANT_ID,
)
from ..models import Action, ActionSchema
from ..tracing._traced import traced
from ._base_service import BaseService


def _create_spec(
    data: Optional[Dict[str, Any]],
    action_schema: Optional[ActionSchema],
    title: str,
    app_key: Optional[str] = None,
    app_version: Optional[int] = 1,
    app_folder_key: Optional[str] = None,
    app_folder_path: Optional[str] = None,
) -> RequestSpec:
    field_list = []
    outcome_list = []
    if action_schema:
        if action_schema.inputs:
            for input_field in action_schema.inputs:
                field_name = input_field.name
                field_list.append(
                    {
                        "Id": input_field.key,
                        "Name": field_name,
                        "Title": field_name,
                        "Type": "Fact",
                        "Value": data.get(field_name, "") if data is not None else "",
                    }
                )
        if action_schema.outputs:
            for output_field in action_schema.outputs:
                field_name = output_field.name
                field_list.append(
                    {
                        "Id": output_field.key,
                        "Name": field_name,
                        "Title": field_name,
                        "Type": "Fact",
                        "Value": "",
                    }
                )
        if action_schema.in_outs:
            for inout_field in action_schema.in_outs:
                field_name = inout_field.name
                field_list.append(
                    {
                        "Id": inout_field.key,
                        "Name": field_name,
                        "Title": field_name,
                        "Type": "Fact",
                        "Value": data.get(field_name, "") if data is not None else "",
                    }
                )
        if action_schema.outcomes:
            for outcome in action_schema.outcomes:
                outcome_list.append(
                    {
                        "Id": action_schema.key,
                        "Name": outcome.name,
                        "Title": outcome.name,
                        "Type": "Action.Http",
                        "IsPrimary": True,
                    }
                )

    return RequestSpec(
        method="POST",
        endpoint=Endpoint("/orchestrator_/tasks/AppTasks/CreateAppTask"),
        content=dumps(
            {
                "appId": app_key,
                "appVersion": app_version,
                "title": title,
                "data": data if data is not None else {},
                "actionableMessageMetaData": {
                    "fieldSet": {
                        "id": str(uuid.uuid4()),
                        "fields": field_list,
                    }
                    if len(field_list) != 0
                    else {},
                    "actionSet": {
                        "id": str(uuid.uuid4()),
                        "actions": outcome_list,
                    }
                    if len(outcome_list) != 0
                    else {},
                }
                if action_schema is not None
                else {},
            }
        ),
        headers=folder_headers(app_folder_key, app_folder_path),
    )


def _retrieve_action_spec(
    action_key: str, app_folder_key: str, app_folder_path: str
) -> RequestSpec:
    return RequestSpec(
        method="GET",
        endpoint=Endpoint("/orchestrator_/tasks/GenericTasks/GetTaskDataByKey"),
        params={"taskKey": action_key},
        headers=folder_headers(app_folder_key, app_folder_path),
    )


def _assign_task_spec(task_key: str, assignee: str) -> RequestSpec:
    return RequestSpec(
        method="POST",
        endpoint=Endpoint(
            "/orchestrator_/odata/Tasks/UiPath.Server.Configuration.OData.AssignTasks"
        ),
        content=dumps(
            {"taskAssignments": [{"taskId": task_key, "UserNameOrEmail": assignee}]}
        ),
    )


def _retrieve_app_key_spec(app_name: str) -> RequestSpec:
    tenant_id = os.getenv(ENV_TENANT_ID, None)
    if not tenant_id:
        raise Exception(f"{ENV_TENANT_ID} env var is not set")
    return RequestSpec(
        method="GET",
        endpoint=Endpoint("/apps_/default/api/v1/default/deployed-action-apps-schemas"),
        params={"search": app_name},
        headers={HEADER_TENANT_ID: tenant_id},
    )


def folder_headers(
    app_folder_key: Optional[str], app_folder_path: Optional[str]
) -> Dict[str, str]:
    headers = {}
    if app_folder_key:
        headers[HEADER_FOLDER_KEY] = app_folder_key
    elif app_folder_path:
        headers[HEADER_FOLDER_PATH] = app_folder_path
    return headers


class ActionsService(FolderContext, BaseService):
    """Service for managing UiPath Actions.

    Actions are task-based automation components that can be integrated into
    applications and processes. They represent discrete units of work that can
    be triggered and monitored through the UiPath API.

    This service provides methods to create and retrieve actions, supporting
    both app-specific and generic actions. It inherits folder context management
    capabilities from FolderContext.

    Reference: https://docs.uipath.com/automation-cloud/docs/actions
    """

    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="actions_create", run_type="uipath")
    async def create_async(
        self,
        title: str,
        data: Optional[Dict[str, Any]] = None,
        *,
        app_name: Optional[str] = None,
        app_key: Optional[str] = None,
        app_folder_path: Optional[str] = None,
        app_folder_key: Optional[str] = None,
        app_version: Optional[int] = 1,
        assignee: Optional[str] = None,
    ) -> Action:
        """Creates a new action asynchronously.

        This method creates a new action task in UiPath Orchestrator. The action can be
        either app-specific (using app_name or app_key) or a generic action.

        Args:
            title: The title of the action
            data: Optional dictionary containing input data for the action
            app_name: The name of the application (if creating an app-specific action)
            app_key: The key of the application (if creating an app-specific action)
            app_folder_path: Optional folder path for the action
            app_folder_key: Optional folder key for the action
            app_version: The version of the application
            assignee: Optional username or email to assign the task to

        Returns:
            Action: The created action object

        Raises:
            Exception: If neither app_name nor app_key is provided for app-specific actions
        """
        app_folder_path = app_folder_path if app_folder_path else self._folder_path

        (key, action_schema) = (
            (app_key, None)
            if app_key
            else await self._get_app_key_and_schema_async(app_name, app_folder_path)
        )
        spec = _create_spec(
            title=title,
            data=data,
            app_key=key,
            app_version=app_version,
            action_schema=action_schema,
            app_folder_key=app_folder_key,
            app_folder_path=app_folder_path,
        )

        response = await self.request_async(
            spec.method, spec.endpoint, content=spec.content, headers=spec.headers
        )
        json_response = response.json()
        if assignee:
            spec = _assign_task_spec(json_response["id"], assignee)
            await self.request_async(spec.method, spec.endpoint, content=spec.content)
        return Action.model_validate(json_response)

    @traced(name="actions_create", run_type="uipath")
    def create(
        self,
        title: str,
        data: Optional[Dict[str, Any]] = None,
        *,
        app_name: Optional[str] = None,
        app_key: Optional[str] = None,
        app_folder_path: Optional[str] = None,
        app_folder_key: Optional[str] = None,
        app_version: Optional[int] = 1,
        assignee: Optional[str] = None,
    ) -> Action:
        """Creates a new action synchronously.

        This method creates a new action task in UiPath Orchestrator. The action can be
        either app-specific (using app_name or app_key) or a generic action.

        Args:
            title: The title of the action
            data: Optional dictionary containing input data for the action
            app_name: The name of the application (if creating an app-specific action)
            app_key: The key of the application (if creating an app-specific action)
            app_folder_path: Optional folder path for the action
            app_folder_key: Optional folder key for the action
            app_version: The version of the application
            assignee: Optional username or email to assign the task to

        Returns:
            Action: The created action object

        Raises:
            Exception: If neither app_name nor app_key is provided for app-specific actions
        """
        app_folder_path = app_folder_path if app_folder_path else self._folder_path

        (key, action_schema) = (
            (app_key, None)
            if app_key
            else self._get_app_key_and_schema(app_name, app_folder_path)
        )
        spec = _create_spec(
            title=title,
            data=data,
            app_key=key,
            app_version=app_version,
            action_schema=action_schema,
            app_folder_key=app_folder_key,
            app_folder_path=app_folder_path,
        )

        response = self.request(
            spec.method, spec.endpoint, content=spec.content, headers=spec.headers
        )
        json_response = response.json()
        if assignee:
            spec = _assign_task_spec(json_response["id"], assignee)
            self.request(spec.method, spec.endpoint, content=spec.content)
        return Action.model_validate(json_response)

    @traced(name="actions_retrieve", run_type="uipath")
    def retrieve(
        self, action_key: str, app_folder_path: str = "", app_folder_key: str = ""
    ) -> Action:
        """Retrieves an action by its key synchronously.

        Args:
            action_key: The unique identifier of the action to retrieve
            app_folder_path: Optional folder path for the action
            app_folder_key: Optional folder key for the action

        Returns:
            Action: The retrieved action object
        """
        spec = _retrieve_action_spec(
            action_key=action_key,
            app_folder_key=app_folder_key,
            app_folder_path=app_folder_path,
        )
        response = self.request(
            spec.method, spec.endpoint, params=spec.params, headers=spec.headers
        )

        return Action.model_validate(response.json())

    @traced(name="actions_retrieve", run_type="uipath")
    async def retrieve_async(
        self, action_key: str, app_folder_path: str = "", app_folder_key: str = ""
    ) -> Action:
        """Retrieves an action by its key asynchronously.

        Args:
            action_key: The unique identifier of the action to retrieve
            app_folder_path: Optional folder path for the action
            app_folder_key: Optional folder key for the action

        Returns:
            Action: The retrieved action object
        """
        spec = _retrieve_action_spec(
            action_key=action_key,
            app_folder_key=app_folder_key,
            app_folder_path=app_folder_path,
        )
        response = await self.request_async(
            spec.method, spec.endpoint, params=spec.params, headers=spec.headers
        )

        return Action.model_validate(response.json())

    async def _get_app_key_and_schema_async(
        self, app_name: Optional[str], app_folder_path: Optional[str]
    ) -> Tuple[str, Optional[ActionSchema]]:
        if not app_name:
            raise Exception("appName or appKey is required")
        spec = _retrieve_app_key_spec(app_name=app_name)

        response = await self.request_async(
            spec.method,
            spec.endpoint,
            params=spec.params,
            headers=spec.headers,
            scoped="org",
        )
        try:
            deployed_app = self._extract_deployed_app(
                response.json()["deployed"], app_folder_path
            )
            action_schema = deployed_app["actionSchema"]
            deployed_app_key = deployed_app["systemName"]
        except (KeyError, IndexError):
            raise Exception("Action app not found") from None
        try:
            return (
                deployed_app_key,
                ActionSchema(
                    key=action_schema["key"],
                    in_outs=action_schema["inOuts"],
                    inputs=action_schema["inputs"],
                    outputs=action_schema["outputs"],
                    outcomes=action_schema["outcomes"],
                ),
            )
        except KeyError:
            raise Exception("Failed to deserialize action schema") from KeyError

    def _get_app_key_and_schema(
        self, app_name: Optional[str], app_folder_path: Optional[str]
    ) -> Tuple[str, Optional[ActionSchema]]:
        if not app_name:
            raise Exception("appName or appKey is required")

        spec = _retrieve_app_key_spec(app_name=app_name)

        response = self.request(
            spec.method,
            spec.endpoint,
            params=spec.params,
            headers=spec.headers,
            scoped="org",
        )

        try:
            deployed_app = self._extract_deployed_app(
                response.json()["deployed"], app_folder_path
            )
            action_schema = deployed_app["actionSchema"]
            deployed_app_key = deployed_app["systemName"]
        except (KeyError, IndexError):
            raise Exception("Action app not found") from None
        try:
            return (
                deployed_app_key,
                ActionSchema(
                    key=action_schema["key"],
                    in_outs=action_schema["inOuts"],
                    inputs=action_schema["inputs"],
                    outputs=action_schema["outputs"],
                    outcomes=action_schema["outcomes"],
                ),
            )
        except KeyError:
            raise Exception("Failed to deserialize action schema") from KeyError

    # should be removed after folder filtering support is added on apps API
    def _extract_deployed_app(
        self, deployed_apps: List[Dict[str, Any]], app_folder_path: Optional[str]
    ) -> Dict[str, Any]:
        if len(deployed_apps) > 1 and not app_folder_path:
            raise Exception("Multiple app schemas found")
        try:
            if app_folder_path:
                return next(
                    app
                    for app in deployed_apps
                    if app["deploymentFolder"]["fullyQualifiedName"] == app_folder_path
                )
            else:
                return next(
                    app
                    for app in deployed_apps
                    if app["deploymentFolder"]["key"] == self._folder_key
                )
        except StopIteration:
            raise KeyError from StopIteration

    @property
    def custom_headers(self) -> Dict[str, str]:
        return self.folder_headers



import importlib
import inspect
import json
import logging
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple

from opentelemetry import trace

from ._utils import _SpanUtils

logger = logging.getLogger(__name__)

_tracer_instance: Optional[trace.Tracer] = None


def get_tracer() -> trace.Tracer:
    """Lazily initializes and returns the tracer instance."""
    global _tracer_instance
    if _tracer_instance is None:
        logger.warning(
            "Initializing tracer instance. This should only be done once per process."
        )
        _tracer_instance = trace.get_tracer(__name__)
    return _tracer_instance


class TracingManager:
    """Static utility class to manage tracing implementations and decorated functions."""

    # Registry to track original functions, decorated functions, and their parameters
    # Each entry is (original_func, decorated_func, params)
    _traced_registry: List[Tuple[Callable[..., Any], Callable[..., Any], Any]] = []

    # Custom tracer implementation
    _custom_tracer_implementation = None  # Custom span provider function
    _current_span_provider: Optional[Callable[[], Any]] = None

    @classmethod
    def get_custom_tracer_implementation(cls):
        """Get the currently set custom tracer implementation."""
        return cls._custom_tracer_implementation

    @classmethod
    def register_current_span_provider(
        cls, current_span_provider: Optional[Callable[[], Any]]
    ):
        """Register a custom current span provider function.

        Args:
            current_span_provider: A function that returns the current span from an external
                                 tracing framework. If None, no custom span parenting will be used.
        """
        cls._current_span_provider = current_span_provider

    @classmethod
    def get_parent_context(cls):
        """Get the parent context using the registered current span provider.

        Returns:
            Context object with the current span set, or None if no provider is registered.
        """
        if cls._current_span_provider is not None:
            try:
                current_span = cls._current_span_provider()
                if current_span is not None:
                    from opentelemetry.trace import set_span_in_context

                    return set_span_in_context(current_span)
            except Exception as e:
                logger.warning(f"Error getting current span from provider: {e}")
                return None
        return None

    @classmethod
    def register_traced_function(cls, original_func, decorated_func, params):
        """Register a function decorated with @traced and its parameters.

        Args:
            original_func: The original function before decoration
            decorated_func: The function after decoration
            params: The parameters used for tracing
        """
        cls._traced_registry.append((original_func, decorated_func, params))

    @classmethod
    def reapply_traced_decorator(cls, tracer_implementation):
        """Reapply a different tracer implementation to all functions previously decorated with @traced.

        Args:
            tracer_implementation: A function that takes the same parameters as _opentelemetry_traced
                                 and returns a decorator. If None, reverts to default implementation.
        """
        tracer_implementation = tracer_implementation or _opentelemetry_traced
        cls._custom_tracer_implementation = tracer_implementation

        # Work with a copy of the registry to avoid modifying it during iteration
        registry_copy = cls._traced_registry.copy()

        for original_func, decorated_func, params in registry_copy:
            # Apply the new decorator with the same parameters
            supported_params = _get_supported_params(tracer_implementation, params)
            new_decorated_func = tracer_implementation(**supported_params)(
                original_func
            )

            logger.debug(
                f"Reapplying decorator to {original_func.__name__}, from {decorated_func.__name__}"
            )

            # If this is a method on a class, we need to update the class
            if hasattr(original_func, "__self__") and hasattr(
                original_func, "__func__"
            ):
                setattr(
                    original_func.__self__.__class__,
                    original_func.__name__,
                    new_decorated_func.__get__(
                        original_func.__self__, original_func.__self__.__class__
                    ),
                )
            else:
                # Replace the function in its module
                if hasattr(original_func, "__module__") and hasattr(
                    original_func, "__qualname__"
                ):
                    try:
                        module = importlib.import_module(original_func.__module__)
                        parts = original_func.__qualname__.split(".")

                        # Handle nested objects
                        obj = module
                        for part in parts[:-1]:
                            obj = getattr(obj, part)

                        setattr(obj, parts[-1], new_decorated_func)

                        # Update the registry entry for this function
                        # Find the index and replace with updated entry
                        for i, (orig, _dec, _p) in enumerate(cls._traced_registry):
                            if orig is original_func:
                                cls._traced_registry[i] = (
                                    original_func,
                                    new_decorated_func,
                                    params,
                                )
                                break
                    except (ImportError, AttributeError) as e:
                        # Log the error but continue processing other functions
                        logger.warning(f"Error reapplying decorator: {e}")
                        continue


def _default_input_processor(inputs):
    """Default input processor that doesn't log any actual input data."""
    return {"redacted": "Input data not logged for privacy/security"}


def _default_output_processor(outputs):
    """Default output processor that doesn't log any actual output data."""
    return {"redacted": "Output data not logged for privacy/security"}


def wait_for_tracers():
    """Wait for all tracers to finish."""
    trace.get_tracer_provider().shutdown()  # type: ignore


def _opentelemetry_traced(
    name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
):
    """Default tracer implementation using OpenTelemetry."""

    def decorator(func):
        trace_name = name if name is not None else func.__name__

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            context = TracingManager.get_parent_context()

            with get_tracer().start_as_current_span(
                trace_name, context=context
            ) as span:
                default_span_type = "function_call_sync"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                try:
                    result = func(*args, **kwargs)
                    # Process output if processor is provided
                    output = result
                    if output_processor is not None:
                        output = output_processor(result)
                    span.set_attribute("output", json.dumps(output, default=str))
                    return result
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            context = TracingManager.get_parent_context()

            with get_tracer().start_as_current_span(
                trace_name, context=context
            ) as span:
                default_span_type = "function_call_async"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                try:
                    result = await func(*args, **kwargs)
                    # Process output if processor is provided
                    output = result
                    if output_processor is not None:
                        output = output_processor(result)
                    span.set_attribute("output", json.dumps(output, default=str))
                    return result
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        def generator_wrapper(*args, **kwargs):
            context = TracingManager.get_parent_context()

            with get_tracer().start_as_current_span(
                trace_name, context=context
            ) as span:
                span.get_span_context()
                default_span_type = "function_call_generator_sync"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                outputs = []
                try:
                    for item in func(*args, **kwargs):
                        outputs.append(item)
                        span.add_event(f"Yielded: {item}")  # Add event for each yield
                        yield item

                    # Process output if processor is provided
                    output_to_record = outputs
                    if output_processor is not None:
                        output_to_record = output_processor(outputs)
                    span.set_attribute(
                        "output", json.dumps(output_to_record, default=str)
                    )
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        async def async_generator_wrapper(*args, **kwargs):
            context = TracingManager.get_parent_context()

            with get_tracer().start_as_current_span(
                trace_name, context=context
            ) as span:
                default_span_type = "function_call_generator_async"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                outputs = []
                try:
                    async for item in func(*args, **kwargs):
                        outputs.append(item)
                        span.add_event(f"Yielded: {item}")  # Add event for each yield
                        yield item

                    # Process output if processor is provided
                    output_to_record = outputs
                    if output_processor is not None:
                        output_to_record = output_processor(outputs)
                    span.set_attribute(
                        "output", json.dumps(output_to_record, default=str)
                    )
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        elif inspect.isgeneratorfunction(func):
            return generator_wrapper
        elif inspect.isasyncgenfunction(func):
            return async_generator_wrapper
        else:
            return sync_wrapper

    return decorator


def _get_supported_params(tracer_impl, params):
    """Extract the parameters supported by the tracer implementation.

    Args:
        tracer_impl: The tracer implementation function or callable
        params: Dictionary of parameters to check

    Returns:
        Dictionary containing only parameters supported by the tracer implementation
    """
    supported_params = {}
    if hasattr(tracer_impl, "__code__"):
        # For regular functions
        impl_signature = inspect.signature(tracer_impl)
        for param_name, param_value in params.items():
            if param_name in impl_signature.parameters and param_value is not None:
                supported_params[param_name] = param_value
    elif callable(tracer_impl):
        # For callable objects
        impl_signature = inspect.signature(tracer_impl.__call__)
        for param_name, param_value in params.items():
            if param_name in impl_signature.parameters and param_value is not None:
                supported_params[param_name] = param_value
    else:
        # If we can't inspect, pass all parameters and let the function handle it
        supported_params = params

    return supported_params


def traced(
    name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
    hide_input: bool = False,
    hide_output: bool = False,
):
    """Decorator that will trace function invocations.

    Args:
        run_type: Optional string to categorize the run type
        span_type: Optional string to categorize the span type
        input_processor: Optional function to process function inputs before recording
            Should accept a dictionary of inputs and return a processed dictionary
        output_processor: Optional function to process function outputs before recording
            Should accept the function output and return a processed value
        hide_input: If True, don't log any input data
        hide_output: If True, don't log any output data
    """
    # Apply default processors selectively based on hide flags
    if hide_input:
        input_processor = _default_input_processor
    if hide_output:
        output_processor = _default_output_processor

    # Store the parameters for later reapplication
    params = {
        "name": name,
        "run_type": run_type,
        "span_type": span_type,
        "input_processor": input_processor,
        "output_processor": output_processor,
    }

    # Check for custom implementation first
    custom_implementation = TracingManager.get_custom_tracer_implementation()
    tracer_impl: Any = (
        custom_implementation if custom_implementation else _opentelemetry_traced
    )

    def decorator(func):
        # Check which parameters are supported by the tracer_impl
        supported_params = _get_supported_params(tracer_impl, params)

        # Decorate the function with only supported parameters
        decorated_func = tracer_impl(**supported_params)(func)

        # Register both original and decorated function with parameters
        TracingManager.register_traced_function(func, decorated_func, params)
        return decorated_func

    return decorator



# type: ignore
import json
import os
import socket
import webbrowser

import click
from dotenv import load_dotenv

from ..telemetry import track
from ._auth._auth_server import HTTPServer
from ._auth._client_credentials import ClientCredentialsService
from ._auth._oidc_utils import get_auth_config, get_auth_url
from ._auth._portal_service import PortalService, select_tenant
from ._auth._utils import update_auth_file, update_env_file
from ._utils._common import environment_options
from ._utils._console import ConsoleLogger

load_dotenv(override=True)
console = ConsoleLogger()


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("localhost", port))
            s.close()
            return False
        except socket.error:
            return True


def set_port():
    auth_config = get_auth_config()
    port = auth_config.get("port", 8104)
    port_option_one = auth_config.get("portOptionOne", 8104)
    port_option_two = auth_config.get("portOptionTwo", 8055)
    port_option_three = auth_config.get("portOptionThree", 42042)
    if is_port_in_use(port):
        if is_port_in_use(port_option_one):
            if is_port_in_use(port_option_two):
                if is_port_in_use(port_option_three):
                    console.error(
                        "All configured ports are in use. Please close applications using ports or configure different ports."
                    )
                else:
                    port = port_option_three
            else:
                port = port_option_two
        else:
            port = port_option_one
    auth_config["port"] = port
    with open(
        os.path.join(os.path.dirname(__file__), "..", "auth_config.json"), "w"
    ) as f:
        json.dump(auth_config, f)


@click.command()
@environment_options
@click.option(
    "-f",
    "--force",
    is_flag=True,
    required=False,
    help="Force new token",
)
@click.option(
    "--client-id",
    required=False,
    help="Client ID for client credentials authentication (unattended mode)",
)
@click.option(
    "--client-secret",
    required=False,
    help="Client secret for client credentials authentication (unattended mode)",
)
@click.option(
    "--base-url",
    required=False,
    help="Base URL for the UiPath tenant instance (required for client credentials)",
)
@track
def auth(
    domain,
    force: None | bool = False,
    client_id: str = None,
    client_secret: str = None,
    base_url: str = None,
):
    """Authenticate with UiPath Cloud Platform.

    Interactive mode (default): Opens browser for OAuth authentication.
    Unattended mode: Use --client-id, --client-secret and --base-url for client credentials flow.

    Network options:
    - Set HTTP_PROXY/HTTPS_PROXY/NO_PROXY environment variables for proxy configuration
    - Set REQUESTS_CA_BUNDLE to specify a custom CA bundle for SSL verification
    - Set UIPATH_DISABLE_SSL_VERIFY to disable SSL verification (not recommended)
    """
    # Check if client credentials are provided for unattended authentication
    if client_id and client_secret:
        if not base_url:
            console.error(
                "--base-url is required when using client credentials authentication."
            )
            return

        with console.spinner("Authenticating with client credentials ..."):
            credentials_service = ClientCredentialsService(domain)

            # If base_url is provided, extract domain from it to override the CLI domain parameter
            if base_url:
                extracted_domain = credentials_service.extract_domain_from_base_url(
                    base_url
                )
                credentials_service.domain = extracted_domain

            token_data = credentials_service.authenticate(client_id, client_secret)

            if token_data:
                credentials_service.setup_environment(token_data, base_url)
                console.success(
                    "Client credentials authentication successful.",
                )
            else:
                console.error(
                    "Client credentials authentication failed. Please check your credentials.",
                )
        return

    # Interactive authentication flow (existing logic)
    with console.spinner("Authenticating with UiPath ..."):
        with PortalService(domain) as portal_service:
            if not force:
                if (
                    os.getenv("UIPATH_URL")
                    and os.getenv("UIPATH_TENANT_ID")
                    and os.getenv("UIPATH_ORGANIZATION_ID")
                ):
                    try:
                        portal_service.ensure_valid_token()
                        console.success(
                            "Authentication successful.",
                        )
                        return
                    except Exception:
                        console.info(
                            "Authentication token is invalid. Please reauthenticate.",
                        )

            auth_url, code_verifier, state = get_auth_url(domain)

            webbrowser.open(auth_url, 1)
            auth_config = get_auth_config()

            console.link(
                "If a browser window did not open, please open the following URL in your browser:",
                auth_url,
            )

            server = HTTPServer(port=auth_config["port"])
            token_data = server.start(state, code_verifier, domain)

            if token_data:
                portal_service.update_token_data(token_data)
                update_auth_file(token_data)
                access_token = token_data["access_token"]
                update_env_file({"UIPATH_ACCESS_TOKEN": access_token})

                tenants_and_organizations = (
                    portal_service.get_tenants_and_organizations()
                )
                base_url = select_tenant(domain, tenants_and_organizations)
                try:
                    portal_service.post_auth(base_url)
                    console.success(
                        "Authentication successful.",
                    )
                except Exception:
                    console.error(
                        "Could not prepare the environment. Please try again.",
                    )
            else:
                console.error(
                    "Authentication failed. Please try again.",
                )



import json
import os
import urllib.parse
from unittest.mock import patch

import pytest
from click.testing import CliRunner
from pytest_httpx import HTTPXMock

from uipath._cli.cli_publish import publish  # type: ignore


@pytest.fixture
def mock_feeds_response() -> list[tuple[str, str]]:
    """Fixture to provide mock feeds response."""
    return [
        ("Tenant Feed", "tenant_feed_id"),
        ("other Feed", "other_feed_id"),
        ("My workspace Feed", "my_workspace_feed_id"),
    ]


def _create_env_file(mock_env_vars: dict[str, str]):
    """Create the environment file."""
    with open(".env", "w") as f:
        for key, value in mock_env_vars.items():
            f.write(f"{key}={value}\n")


class TestPublish:
    def test_publish_no_env(
        self,
        runner: CliRunner,
        temp_dir: str,
    ) -> None:
        """Test publish command when no env file exists."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            result = runner.invoke(publish)
            assert result.exit_code == 1
            assert "Missing required environment variables." in result.output
            assert "UIPATH_URL" in result.output
            assert "UIPATH_ACCESS_TOKEN" in result.output

    def test_publish_no_package(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        mock_feeds_response: list[tuple[str, str]],
    ) -> None:
        """Test publish command when no package exists."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            with (
                patch(
                    "uipath._cli.cli_publish.get_available_feeds"
                ) as get_available_feeds_mock,
                patch("uipath._cli.cli_publish.console.prompt") as mock_prompt,
            ):
                get_available_feeds_mock.return_value = mock_feeds_response
                mock_prompt.return_value = 0
                _create_env_file(mock_env_vars)
                os.makedirs(".uipath")
                result = runner.invoke(publish)
                assert result.exit_code == 1
                assert (
                    "No .nupkg files found. Please run `uipath pack` first."
                    in result.output
                )

    def test_publish_invalid_feed(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        mock_feeds_response: list[tuple[str, str]],
    ) -> None:
        """Test publish command with invalid feed selections."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            with (
                patch(
                    "uipath._cli.cli_publish.get_available_feeds"
                ) as get_available_feeds_mock,
                patch("uipath._cli.cli_publish.console.prompt") as mock_prompt,
            ):
                get_available_feeds_mock.return_value = mock_feeds_response
                _create_env_file(mock_env_vars)
                # negative value
                mock_prompt.return_value = -1
                result = runner.invoke(publish)
                assert result.exit_code == 1
                assert "Invalid feed selected" in result.output
                # invalid type
                mock_prompt.return_value = "string"
                result = runner.invoke(publish)
                assert result.exit_code == 1
                assert type(result.exception) is TypeError
                # index error
                mock_prompt.return_value = len(mock_feeds_response) + 1
                result = runner.invoke(publish)
                assert result.exit_code == 1
                assert "Invalid feed selected" in result.output

    def test_publish_error(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        httpx_mock: HTTPXMock,
    ) -> None:
        """Test successful publish to tenant feed."""
        base_url = mock_env_vars.get("UIPATH_URL")
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()",
            status_code=401,
        )

        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(os.path.join(".uipath", "test.1.0.0.nupkg"), "wb") as f:
                f.write(b"dummy package content")

            _create_env_file(mock_env_vars)

            result = runner.invoke(publish, ["--tenant"])

            assert result.exit_code == 1
            assert "Failed to publish package. Status code: 401" in result.output

    def test_publish_tenant_feed_success(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        httpx_mock: HTTPXMock,
    ) -> None:
        """Test successful publish to tenant feed."""
        base_url = mock_env_vars.get("UIPATH_URL")
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()",
            status_code=200,
        )

        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(os.path.join(".uipath", "test.1.0.0.nupkg"), "wb") as f:
                f.write(b"dummy package content")

            _create_env_file(mock_env_vars)

            result = runner.invoke(publish, ["--tenant"])

            assert result.exit_code == 0
            assert "Package published successfully!" in result.output

    def test_publish_folder_feed_success(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        httpx_mock: HTTPXMock,
        mock_feeds_response: list[tuple[str, str]],
    ) -> None:
        """Test successful publish to a folder feed."""
        base_url = mock_env_vars.get("UIPATH_URL")
        folder_feed_name, folder_feed_id = mock_feeds_response[1]  # Unpack the tuple

        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()?feedId={folder_feed_id}",
            status_code=200,
        )

        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(os.path.join(".uipath", "test.1.0.0.nupkg"), "wb") as f:
                f.write(b"dummy package content")

            with (
                patch("uipath._cli.cli_publish.console.prompt") as mock_prompt,
                patch("uipath._cli.cli_publish.get_available_feeds") as mock_feeds,
                patch(
                    "uipath._cli.cli_publish.get_personal_workspace_info"
                ) as mock_workspace,
            ):
                mock_prompt.return_value = 1  # Select second feed (index 1)
                mock_feeds.return_value = mock_feeds_response
                mock_workspace.return_value = (
                    "different-feed-id",
                    "folder-id",
                )  # Different from our target feed
                _create_env_file(mock_env_vars)

                result = runner.invoke(publish)

                assert result.exit_code == 0
                assert "Package published successfully!" in result.output

                # Verify the methods were called correctly
                mock_prompt.assert_called_once_with("Select feed number", type=int)
                mock_feeds.assert_called_once()
                mock_workspace.assert_called_once_with(
                    base_url, mock_env_vars.get("UIPATH_ACCESS_TOKEN")
                )

    def test_publish_personal_workspace_not_found(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
    ) -> None:
        """Test publish to personal workspace when workspace not found."""
        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(os.path.join(".uipath", "test.1.0.0.nupkg"), "wb") as f:
                f.write(b"dummy package content")

            with patch(
                "uipath._cli.cli_publish.get_personal_workspace_info"
            ) as mock_workspace:
                _create_env_file(mock_env_vars)
                mock_workspace.return_value = (None, None)

                result = runner.invoke(publish, ["--my-workspace"])

                assert result.exit_code == 1
                assert "No personal workspace found for user" in result.output

    def test_publish_my_workspace_feed_success(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        httpx_mock: HTTPXMock,
    ) -> None:
        """Test successful publish to tenant feed."""
        my_workspace_feed_id = "042e669a-c95f-46a3-87b0-9e5a98d7cf8a"
        release_id = "123"
        package_name = "mock_package_name"
        package_version = "mock__package_version"
        my_workspace_folder_id = "123"
        upload_package_response_data = {
            "value": [
                {"Body": f'{{"Id":"{package_name}","Version":"{package_version}"}}'}
            ]
        }
        list_release_response_data = {
            "value": [
                {
                    "ProcessVersion": package_version,
                    "Id": release_id,
                    "Key": "9d17b737-1283-4ebe-b1f5-7d88967b94e4",
                }
            ]
        }
        personal_workspace_response_data = {
            "PersonalWorskpaceFeedId": my_workspace_feed_id,
            "PersonalWorkspace": {"Id": my_workspace_folder_id},
        }
        base_url = mock_env_vars.get("UIPATH_URL")
        # upload package
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()?feedId={my_workspace_feed_id}",
            status_code=200,
            text=json.dumps(upload_package_response_data),
        )
        odata_top_filter = 25
        # get release info
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Releases/UiPath.Server.Configuration.OData.ListReleases?$select=Id,Key,ProcessVersion&$top={odata_top_filter}&$filter=ProcessKey%20eq%20%27{urllib.parse.quote(package_name)}%27",
            status_code=200,
            text=json.dumps(list_release_response_data),
        )
        # get personal workspace info
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Users/UiPath.Server.Configuration.OData.GetCurrentUserExtended?$expand=PersonalWorkspace",
            status_code=200,
            text=json.dumps(personal_workspace_response_data),
        )

        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(
                os.path.join(".uipath", f"{package_name}.{package_version}.nupkg"), "wb"
            ) as f:
                f.write(b"dummy package content")

            _create_env_file(mock_env_vars)

            result = runner.invoke(publish, ["--my-workspace"])
            expected_release_url = f"{base_url}/orchestrator_/processes/{release_id}/edit?fid={my_workspace_folder_id}"
            assert result.exit_code == 0
            assert "Package published successfully!" in result.output
            assert expected_release_url in result.output
            assert (
                "Use the link above to configure any environment variables"
                in result.output
            )

    def test_publish_my_workspace_feed_without_flag(
        self,
        runner: CliRunner,
        temp_dir: str,
        mock_env_vars: dict[str, str],
        httpx_mock: HTTPXMock,
    ) -> None:
        """Test successful publish to tenant feed."""
        my_workspace_feed_id = "042e669a-c95f-46a3-87b0-9e5a98d7cf8a"
        release_id = "123"
        package_name = "mock_package_name"
        package_version = "mock__package_version"
        my_workspace_folder_id = "123"
        upload_package_response_data = {
            "value": [
                {"Body": f'{{"Id":"{package_name}","Version":"{package_version}"}}'}
            ]
        }
        list_release_response_data = {
            "value": [
                {
                    "ProcessVersion": package_version,
                    "Id": release_id,
                    "Key": "9d17b737-1283-4ebe-b1f5-7d88967b94e4",
                }
            ]
        }
        personal_workspace_response_data = {
            "PersonalWorskpaceFeedId": my_workspace_feed_id,
            "PersonalWorkspace": {"Id": my_workspace_folder_id},
        }
        feeds_response_data = [
            {
                "name": "Tenant Feed",
                "id": "3d732cad-1406-4e58-b2c9-d18ef257f100",
                "purpose": "Processes",
            },
            {
                "name": "other Feed",
                "id": "e6d5a7d5-6bdd-4263-ac48-258a093bca01",
                "purpose": "Processes",
            },
            {
                "name": "My workspace Feed",
                "id": my_workspace_feed_id,
                "purpose": "Processes",
            },
            {
                "name": "another feed",
                "id": "e6d5a7d5-6bdd-4263-ac48-258a093bca01",
                "purpose": "other_purpose",
            },
        ]
        base_url = mock_env_vars.get("UIPATH_URL")
        # upload package
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()?feedId={my_workspace_feed_id}",
            status_code=200,
            text=json.dumps(upload_package_response_data),
        )
        odata_top_filter = 25
        # get release info
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Releases/UiPath.Server.Configuration.OData.ListReleases?$select=Id,Key,ProcessVersion&$top={odata_top_filter}&$filter=ProcessKey%20eq%20%27{urllib.parse.quote(package_name)}%27",
            status_code=200,
            text=json.dumps(list_release_response_data),
        )
        # get personal workspace info
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/odata/Users/UiPath.Server.Configuration.OData.GetCurrentUserExtended?$expand=PersonalWorkspace",
            status_code=200,
            text=json.dumps(personal_workspace_response_data),
        )
        # get available feeds
        httpx_mock.add_response(
            url=f"{base_url}/orchestrator_/api/PackageFeeds/GetFeeds",
            status_code=200,
            text=json.dumps(feeds_response_data),
        )

        with runner.isolated_filesystem(temp_dir=temp_dir):
            os.makedirs(".uipath")
            # Create dummy package
            with open(
                os.path.join(".uipath", f"{package_name}.{package_version}.nupkg"), "wb"
            ) as f:
                f.write(b"dummy package content")

            with patch("uipath._cli.cli_publish.console.prompt") as mock_prompt:
                # return my workspace index
                mock_prompt.return_value = 2
                _create_env_file(mock_env_vars)

                result = runner.invoke(publish)
                expected_release_url = f"{base_url}/orchestrator_/processes/{release_id}/edit?fid={my_workspace_folder_id}"
                assert result.exit_code == 0
                assert "Package published successfully!" in result.output
                assert expected_release_url in result.output
                assert (
                    "Use the link above to configure any environment variables"
                    in result.output
                )



import json
from typing import Any, Dict, List, Optional

from .._config import Config
from .._execution_context import ExecutionContext
from .._utils import Endpoint
from ..models.llm_gateway import (
    ChatCompletion,
    SpecificToolChoice,
    TextEmbedding,
    ToolChoice,
    ToolDefinition,
    UsageInfo,
)
from ..tracing._traced import traced
from ._base_service import BaseService

# Common constants
API_VERSION = "2024-10-21"
NORMALIZED_API_VERSION = "2024-08-01-preview"

# Common headers
DEFAULT_LLM_HEADERS = {
    "X-UIPATH-STREAMING-ENABLED": "false",
    "X-UiPath-LlmGateway-RequestingProduct": "uipath-python-sdk",
    "X-UiPath-LlmGateway-RequestingFeature": "langgraph-agent",
}


class ChatModels(object):
    gpt_4 = "gpt-4"
    gpt_4_1106_Preview = "gpt-4-1106-Preview"
    gpt_4_32k = "gpt-4-32k"
    gpt_4_turbo_2024_04_09 = "gpt-4-turbo-2024-04-09"
    gpt_4_vision_preview = "gpt-4-vision-preview"
    gpt_4o_2024_05_13 = "gpt-4o-2024-05-13"
    gpt_4o_2024_08_06 = "gpt-4o-2024-08-06"
    gpt_4o_mini_2024_07_18 = "gpt-4o-mini-2024-07-18"
    o3_mini = "o3-mini-2025-01-31"


class EmbeddingModels(object):
    text_embedding_3_large = "text-embedding-3-large"
    text_embedding_ada_002 = "text-embedding-ada-002"


API_VERSION = "2024-10-21"
NORMALIZED_API_VERSION = "2024-08-01-preview"


class UiPathOpenAIService(BaseService):
    """Service calling llm gateway service."""

    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="llm_embeddings_usage", run_type="uipath")
    async def embeddings_usage(
        self, input: str, embedding_model: str = EmbeddingModels.text_embedding_ada_002
    ):
        """Embedd the input text using llm gateway service.

        Args:
            input (str): The input text to embedd.
            embedding_model (str, optional): The embedding model to use. Defaults to text-embedding-ada-002.

        Returns:
            EmbeddingUsageInfo: The embedding usage information.
        """
        endpoint = Endpoint(
            f"/llmgateway_/openai/deployments/{embedding_model}/embeddings/usage"
        )

        response = await self.request_async(
            "POST",
            endpoint,
            content=json.dumps({"input": input}),
            params={"api-version": API_VERSION},
            headers=DEFAULT_LLM_HEADERS,
        )

        return UsageInfo.model_validate(response.json())

    @traced(name="llm_embeddings", run_type="uipath")
    async def embeddings(
        self, input: str, embedding_model: str = EmbeddingModels.text_embedding_ada_002
    ):
        """Embed the input text using llm gateway service.

        Args:
            input (str): The input text to embed.

        Returns:
            TextEmbedding: The embedding response.
        """
        endpoint = Endpoint(
            f"/llmgateway_/openai/deployments/{embedding_model}/embeddings"
        )

        response = await self.request_async(
            "POST",
            endpoint,
            content=json.dumps({"input": input}),
            params={"api-version": API_VERSION},
            headers=DEFAULT_LLM_HEADERS,
        )

        return TextEmbedding.model_validate(response.json())

    @traced(name="llm_chat_completions", run_type="uipath")
    async def chat_completions(
        self,
        messages: List[Dict[str, str]],
        model: str = ChatModels.gpt_4o_mini_2024_07_18,
        max_tokens: int = 50,
        temperature: float = 0,
    ):
        """Get chat completions using llm gateway service.

        Args:
            messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content' keys.
                The supported roles are 'system', 'user', and 'assistant'.

        Example:
                ```
                [
                    {"role": "system", "content": "You are a helpful Python programming assistant."},
                    {"role": "user", "content": "How do I read a file in Python?"},
                    {"role": "assistant", "content": "You can use the built-in open() function."},
                    {"role": "user", "content": "Can you show an example?"}
                ]
                ```
                The conversation history can be included to provide context to the model.
            model (str, optional): The model to use for chat completion. Defaults to ChatModels.gpt_4o_mini_2024_07_18.
            max_tokens (int, optional): Maximum number of tokens to generate. Defaults to 50.
            temperature (float, optional): Temperature for sampling, between 0 and 1.
                Lower values make output more deterministic. Defaults to 0.

        Returns:
            ChatCompletion: The chat completion response.
        """
        endpoint = Endpoint(f"/llmgateway_/openai/deployments/{model}/chat/completions")

        request_body = {
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
        }

        response = await self.request_async(
            "POST",
            endpoint,
            content=json.dumps(request_body),
            params={"api-version": API_VERSION},
            headers=DEFAULT_LLM_HEADERS,
        )

        return ChatCompletion.model_validate(response.json())

    @traced(name="llm_chat_completions_usage", run_type="uipath")
    async def chat_completions_usage(
        self,
        messages: List[Dict[str, str]],
        model: str = ChatModels.gpt_4o_mini_2024_07_18,
        max_tokens: int = 50,
        temperature: float = 0,
    ):
        """Get chat completions usage using llm gateway service.

        Args:
            messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content' keys.
                The supported roles are 'system', 'user', and 'assistant'.

        Example:
                ```
                [
                    {"role": "system", "content": "You are a helpful Python programming assistant."},
                    {"role": "user", "content": "How do I read a file in Python?"},
                    {"role": "assistant", "content": "You can use the built-in open() function."},
                    {"role": "user", "content": "Can you show an example?"}
                ]
                ```
                The conversation history can be included to provide context to the model.
            model (str, optional): The model to use for chat completion. Defaults to ChatModels.gpt_4o_mini_2024_07_18.
            max_tokens (int, optional): Maximum number of tokens to generate. Defaults to 50.
            temperature (float, optional): Temperature for sampling, between 0 and 1.
                Lower values make output more deterministic. Defaults to 0.

        Returns:
            ChatCompletion: The chat completion usage response.
        """
        endpoint = Endpoint(
            f"/llmgateway_/openai/deployments/{model}/chat/completions/usage"
        )

        request_body = {
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
        }

        response = await self.request_async(
            "POST",
            endpoint,
            content=json.dumps(request_body),
            params={"api-version": API_VERSION},
            headers=DEFAULT_LLM_HEADERS,
        )

        return UsageInfo.model_validate(response.json())


class UiPathLlmChatService(BaseService):
    """Service for calling UiPath's normalized LLM Gateway API."""

    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="llm_chat_completions", run_type="uipath")
    async def chat_completions(
        self,
        messages: List[Dict[str, str]],
        model: str = ChatModels.gpt_4o_mini_2024_07_18,
        max_tokens: int = 250,
        temperature: float = 0,
        n: int = 1,
        frequency_penalty: float = 0,
        presence_penalty: float = 0,
        top_p: float = 1,
        tools: Optional[List[ToolDefinition]] = None,
        tool_choice: Optional[ToolChoice] = None,
    ):
        """Get chat completions using UiPath's normalized LLM Gateway API.

        Args:
            messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content' keys.
                The supported roles are 'system', 'user', and 'assistant'.
            model (str, optional): The model to use for chat completion. Defaults to ChatModels.gpt_4o_mini_2024_07_18.
            max_tokens (int, optional): Maximum number of tokens to generate. Defaults to 250.
            temperature (float, optional): Temperature for sampling, between 0 and 1.
                Lower values make output more deterministic. Defaults to 0.
            n (int, optional): Number of chat completion choices to generate. Defaults to 1.
            frequency_penalty (float, optional): Penalty for token frequency. Defaults to 0.
            presence_penalty (float, optional): Penalty for token presence. Defaults to 0.
            top_p (float, optional): Nucleus sampling parameter. Defaults to 1.
            tools (Optional[List[ToolDefinition]], optional): List of tool definitions. Defaults to None.
            tool_choice (Optional[ToolChoice], optional): Tool choice configuration.
                Can be "auto", "none", an AutoToolChoice, a RequiredToolChoice, or a SpecificToolChoice. Defaults to None.

        Returns:
            ChatCompletion: The chat completion response.
        """
        endpoint = Endpoint("/llmgateway_/api/chat/completions")

        request_body = {
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "n": n,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "top_p": top_p,
        }

        # Add tools if provided - convert to UiPath format
        if tools:
            request_body["tools"] = [
                self._convert_tool_to_uipath_format(tool) for tool in tools
            ]

        # Handle tool_choice
        if tool_choice:
            if isinstance(tool_choice, str):
                request_body["tool_choice"] = tool_choice
            elif isinstance(tool_choice, SpecificToolChoice):
                request_body["tool_choice"] = {"type": "tool", "name": tool_choice.name}
            else:
                request_body["tool_choice"] = tool_choice.model_dump()

        # Use default headers but update with normalized API specific headers
        headers = {
            **DEFAULT_LLM_HEADERS,
            "X-UiPath-LlmGateway-NormalizedApi-ModelName": model,
        }

        response = await self.request_async(
            "POST",
            endpoint,
            content=json.dumps(request_body),
            params={"api-version": NORMALIZED_API_VERSION},
            headers=headers,
        )

        return ChatCompletion.model_validate(response.json())

    def _convert_tool_to_uipath_format(self, tool: ToolDefinition) -> Dict[str, Any]:
        """Convert an OpenAI-style tool definition directly to UiPath API format."""
        parameters = {
            "type": tool.function.parameters.type,
            "properties": {
                name: {
                    "type": prop.type,
                    **({"description": prop.description} if prop.description else {}),
                    **({"enum": prop.enum} if prop.enum else {}),
                }
                for name, prop in tool.function.parameters.properties.items()
            },
        }

        if tool.function.parameters.required:
            parameters["required"] = tool.function.parameters.required

        return {
            "name": tool.function.name,
            "description": tool.function.description,
            "parameters": parameters,
        }



import os

import httpx
import pytest

from uipath._config import Config
from uipath._execution_context import ExecutionContext
from uipath._services.llm_gateway_service import (
    ChatModels,
    EmbeddingModels,
    UiPathOpenAIService,
)


def get_env_var(name: str) -> str:
    """Get environment variable or skip test if not present."""
    value = os.environ.get(name)
    if value is None:
        pytest.skip(f"Environment variable {name} is not set")
    return value


def get_access_token() -> str:
    try:
        client_id = get_env_var("UIPATH_CLIENT_ID")
        client_secret = get_env_var("UIPATH_CLIENT_SECRET")
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        url = f"{get_env_var('UIPATH_BASE_URL')}/identity_/connect/token"
        response = httpx.post(url, data=payload, headers=headers)
        json = response.json()
        token = json.get("access_token")

        return token
    except Exception:
        pytest.skip("Failed to get access token. Check your credentials.")


class TestLLMIntegration:
    @pytest.fixture
    def llm_service(self):
        """Create an OpenAIService instance with environment variables."""
        # skip tests on CI, only run locally
        pytest.skip("Failed to get access token. Check your credentials.")

        base_url = get_env_var("UIPATH_URL")
        api_key = get_access_token()

        config = Config(base_url=base_url, secret=api_key)
        execution_context = ExecutionContext()
        return UiPathOpenAIService(config=config, execution_context=execution_context)

    @pytest.mark.asyncio
    async def test_embeddings_real(self, llm_service):
        """Test the embeddings function with a real API call."""
        input_text = "This is a test for embedding a sentence."

        # Make the actual API call
        result = await llm_service.embeddings(input=input_text)

        # Validate the response
        assert result is not None
        assert hasattr(result, "data")
        assert len(result.data) > 0
        assert hasattr(result.data[0], "embedding")
        assert len(result.data[0].embedding) > 0
        assert hasattr(result, "model")
        assert hasattr(result, "usage")
        assert result.usage.prompt_tokens > 0

    @pytest.mark.asyncio
    async def test_embeddings_usage_real(self, llm_service):
        """Test the embeddings_usage function with a real API call."""
        input_text = "Testing the embedding usage endpoint."

        # Make the actual API call
        result = await llm_service.embeddings_usage(input=input_text)

        # Validate the response
        assert result is not None
        assert hasattr(result, "encoding")
        assert hasattr(result, "prompt_tokens")
        assert result.prompt_tokens > 0

    @pytest.mark.asyncio
    async def test_chat_completions_real(self, llm_service):
        """Test the chat_completions function with a real API call."""
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "What is the capital of France?"},
        ]

        # Make the actual API call
        result = await llm_service.chat_completions(
            messages=messages,
            model=ChatModels.gpt_4o_mini_2024_07_18,
            max_tokens=50,
            temperature=0.7,
        )

        # Validate the response
        assert result is not None
        assert hasattr(result, "id")
        assert hasattr(result, "choices")
        assert len(result.choices) > 0
        assert hasattr(result.choices[0], "message")
        assert hasattr(result.choices[0].message, "content")
        assert result.choices[0].message.content.strip() != ""
        assert hasattr(result, "usage")
        assert result.usage.prompt_tokens > 0

    @pytest.mark.asyncio
    async def test_chat_completions_usage_real(self, llm_service):
        """Test the chat_completions_usage function with a real API call."""
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "What is the capital of France?"},
        ]

        # Make the actual API call
        result = await llm_service.chat_completions_usage(
            messages=messages,
            model=ChatModels.gpt_4o_mini_2024_07_18,
            max_tokens=50,
            temperature=0.7,
        )

        # Validate the response
        assert result is not None
        assert hasattr(result, "encoding")
        assert hasattr(result, "prompt_tokens")
        assert result.prompt_tokens > 0
        assert isinstance(result.prompt_tokens, int)
        assert isinstance(result.encoding, str)
        assert len(result.encoding) > 0

    @pytest.mark.asyncio
    async def test_embeddings_with_custom_model_real(self, llm_service):
        """Test the embeddings function with a custom model."""
        input_text = "Testing embeddings with a different model."

        # Make the actual API call with a specific embedding model
        result = await llm_service.embeddings(
            input=input_text, embedding_model=EmbeddingModels.text_embedding_3_large
        )

        # Validate the response
        assert result is not None
        assert hasattr(result, "data")
        assert len(result.data) > 0
        assert hasattr(result.data[0], "embedding")
        assert len(result.data[0].embedding) > 0
        assert result.model == "text-embedding-3-large"
